%{
    #include "du3.h"
    #include "duerr.h"
    #include <assert.h>

    /* semantic value */
    #define YYSTYPE mlc::MlaskalLval

	/* locations */
	#define YYLTYPE unsigned
	#define YYLLOC_DEFAULT(cur, rhs, n)	do { if(n) (cur)=(rhs)[1]; else (cur)=(rhs)[0]; } while(0)

    /* local stuff */
    using namespace mlc;
    
	/* error stuff */
    #define YYERROR_VERBOSE 1
    static void yyerror(YYLTYPE *line, mlc::MlaskalCtx *ctx, const char *m);
    
%}

%token DUTOK_PROGRAM			    /* program */
%token DUTOK_LABEL			    /* label */
%token DUTOK_CONST			    /* const */
%token DUTOK_TYPE			    /* type */
%token DUTOK_VAR			    /* var */
%token DUTOK_BEGIN			    /* begin */
%token DUTOK_END			    /* end */
%token DUTOK_PROCEDURE			    /* procedure */
%token DUTOK_FUNCTION			    /* function */
%token DUTOK_ARRAY			    /* array */
%token DUTOK_OF				    /* of */
%token DUTOK_GOTO			    /* goto */
%token DUTOK_IF				    /* if */
%token DUTOK_THEN			    /* then */
%token DUTOK_ELSE			    /* else */
%token DUTOK_WHILE			    /* while */
%token DUTOK_DO				    /* do */
%token DUTOK_REPEAT			    /* repeat */
%token DUTOK_UNTIL			    /* until */
%token DUTOK_FOR			    /* for */
%token DUTOK_OR				    /* or */
%token DUTOK_NOT			    /* not */
%token DUTOK_RECORD			    /* record */

/* literals */
%token DUTOK_IDENTIFIER			    /* identifier */
%token DUTOK_UINT			    /* unsigned integer */
%token DUTOK_REAL			    /* real number */
%token DUTOK_STRING			    /* string */

/* delimiters */
%token DUTOK_SEMICOLON			    /* ; */
%token DUTOK_DOT			    /* . */
%token DUTOK_COMMA			    /* , */
%token DUTOK_EQ				    /* = */
%token DUTOK_COLON			    /* : */
%token DUTOK_LPAR			    /* ( */
%token DUTOK_RPAR			    /* ) */
%token DUTOK_DOTDOT			    /* .. */
%token DUTOK_LSBRA			    /* [ */
%token DUTOK_RSBRA			    /* ] */
%token DUTOK_ASSIGN			    /* := */

/* grouped operators and keywords */
%token DUTOK_OPER_REL			    /* <, <=, <>, >=, > */
%token DUTOK_OPER_SIGNADD		    /* +, - */
%token DUTOK_OPER_MUL			    /* *, /, div, mod, and */
%token DUTOK_FOR_DIRECTION		    /* to, downto */

%start mlaskal
%error-verbose
%pure-parser
%parse-param {mlc::MlaskalCtx *ctx}
%locations

%%


mlaskal:                DUTOK_PROGRAM DUTOK_IDENTIFIER DUTOK_SEMICOLON program_block DUTOK_DOT
		;

program_block:          labelsBlock constantsBlock typesBlock variablesBlock methodsBlock mainBlock
		;
		
labelsBlock:            /* empty */
                       |DUTOK_LABEL labelsList DUTOK_SEMICOLON
        ;                          		
        
labelsList:             DUTOK_UINT
                       |labelsList DUTOK_COMMA DUTOK_UINT
        ;                               

constantsBlock:         /* empty */
                       |DUTOK_CONST constantsList
        ;                               
        
constantsList:          DUTOK_IDENTIFIER DUTOK_EQ constant DUTOK_SEMICOLON
                       |constantsList DUTOK_IDENTIFIER DUTOK_EQ constant DUTOK_SEMICOLON
        ;        
        
typesBlock:             /* empty */
                       |DUTOK_TYPE typesList
        ;
        
typesList:              DUTOK_IDENTIFIER DUTOK_EQ type DUTOK_SEMICOLON
                       |typesList DUTOK_IDENTIFIER DUTOK_EQ type DUTOK_SEMICOLON
        ;                               
        
variablesBlock:         /* empty */
                       |DUTOK_VAR variablesList
        ;
        
variablesList:          variables DUTOK_COLON type DUTOK_SEMICOLON
                       |variablesList variables DUTOK_COLON type DUTOK_SEMICOLON
        ;  
        
variables:              DUTOK_IDENTIFIER
                       |variables DUTOK_COMMA DUTOK_IDENTIFIER
        ;                                                            
        
methodsBlock:           /* empty */
                       |methodHeader DUTOK_SEMICOLON commonBlock DUTOK_SEMICOLON
                       |methodsBlock methodHeader DUTOK_SEMICOLON commonBlock DUTOK_SEMICOLON
        ;
        
methodHeader:           functionHeader
                       |procedureHeader
        ;                       

mainBlock:              DUTOK_BEGIN commandsList DUTOK_END
        ;
        
commandsList:           command
                       |commandsList DUTOK_SEMICOLON command
        ;
        
commonBlock:            labelsBlock constantsBlock typesBlock variablesBlock mainBlock
        ;
        
procedureHeader:        DUTOK_PROCEDURE DUTOK_IDENTIFIER
                       |DUTOK_PROCEDURE DUTOK_IDENTIFIER DUTOK_LPAR formalParametersList DUTOK_RPAR
        ;
        
functionHeader:         DUTOK_FUNCTION DUTOK_IDENTIFIER DUTOK_COLON scalarTypeIdentifier
                       |DUTOK_FUNCTION DUTOK_IDENTIFIER DUTOK_LPAR formalParametersList DUTOK_RPAR DUTOK_COLON scalarTypeIdentifier
        ;
        
formalParametersList:   variables DUTOK_COLON typeIdentifier
                       |DUTOK_VAR variables DUTOK_COLON typeIdentifier
                       |formalParametersList DUTOK_SEMICOLON variables DUTOK_COLON typeIdentifier
                       |formalParametersList DUTOK_SEMICOLON DUTOK_VAR variables DUTOK_COLON typeIdentifier
        ;            
        
type:                   ordinalType
                       |typeIdentifier
                       |structuredType
        ;
        
ordinalType:            ordinalTypeIdentifier
        ;
        
structuredType:         structuredTypeIdentifier
                       |DUTOK_RECORD recordItemsList DUTOK_SEMICOLON DUTOK_END
                       |DUTOK_RECORD recordItemsList DUTOK_END
                       |DUTOK_RECORD DUTOK_END
        ;
        
recordItemsList:        variables DUTOK_COLON type
                       |recordItemsList DUTOK_SEMICOLON variables DUTOK_COLON type
        ;
        
command:                labelForJump innerCommand
                       |innerCommand
        ;
        
innerCommand:           assignCmd
                       |procedureCallCmd
                       |gotoCmd
                       |blockCmd
                       |conditionCmd
                       |whileCmd
                       |repeatCmd
                       |forCmd
                       |nullCmd
        ;
        
labelForJump:           DUTOK_UINT DUTOK_COLON
        ;
        
assignCmd:              variable DUTOK_ASSIGN expression
                       |functionIdentifier DUTOK_ASSIGN expression
        ;
        
procedureCallCmd:       procedureIdentifier
                       |procedureIdentifier DUTOK_LPAR realParameters DUTOK_RPAR
        ;
        
gotoCmd:                DUTOK_GOTO DUTOK_UINT
        ;
        
blockCmd:               DUTOK_BEGIN commandsList DUTOK_END
        ;
        
conditionCmd:           DUTOK_IF booleanExpression DUTOK_THEN command
                       |DUTOK_IF booleanExpression DUTOK_THEN command DUTOK_ELSE command
        ;
        
whileCmd:               DUTOK_WHILE booleanExpression DUTOK_DO command
        ;
        
repeatCmd:              DUTOK_REPEAT commandsList DUTOK_UNTIL booleanExpression
        ;
        
forCmd:                 DUTOK_FOR ordinalTypedVariableIdentifier DUTOK_ASSIGN ordinalTypedExpression DUTOK_FOR_DIRECTION ordinalTypedExpression DUTOK_DO command
        ;
        
nullCmd:                /* empty */
        ;
        
variable:               variableIdentifier
                       |recordTypedVariable DUTOK_DOT recordItemIdentifier
        ;
        
realParameters:         expression
                       |variable 
                       |realParameters DUTOK_COMMA expression
                       |realParameters DUTOK_COMMA variable
        ;
        
expression:             simpleExpression DUTOK_OPER_REL simpleExpression
                       |simpleExpression DUTOK_EQ simpleExpression
                       |simpleExpression
        ;
        
simpleExpression:       DUTOK_OPER_SIGNADD termsList
                       |termsList
        ;
        
termsList:              termsList DUTOK_OR term
                       |termsList DUTOK_OPER_SIGNADD term
                       |term
        ;
        
term:                   factorsList;
        ;
        
factorsList:            factorsList DUTOK_OPER_MUL factor
                       |factor
        ;
        
factor:                 unsignedConstant
                       |variable
                       |functionIdentifier DUTOK_LPAR realParameters DUTOK_RPAR
                       |functionIdentifier
                       |DUTOK_LPAR expression DUTOK_RPAR
                       |DUTOK_NOT factor
        ;
        
constant:               unsignedConstant
                       |DUTOK_OPER_SIGNADD DUTOK_UINT
                       |DUTOK_OPER_SIGNADD DUTOK_REAL
        ;
        
unsignedConstant:       constantIdentifier
                       |DUTOK_UINT
                       |DUTOK_REAL
                       |DUTOK_STRING
        ;
        
recordTypedVariable:             DUTOK_IDENTIFIER
        ;
        
variableIdentifier:              DUTOK_IDENTIFIER
        ;
        
ordinalTypedVariableIdentifier:  DUTOK_IDENTIFIER
        ;
        
scalarTypeIdentifier:            DUTOK_IDENTIFIER
        ;
        
ordinalTypeIdentifier:           DUTOK_IDENTIFIER
        ;
        
procedureIdentifier:             DUTOK_IDENTIFIER
        ;

constantIdentifier:              DUTOK_IDENTIFIER
        ;
        
recordItemIdentifier:            DUTOK_IDENTIFIER
        ;
        
functionIdentifier:              DUTOK_IDENTIFIER
        ;
        
structuredTypeIdentifier:        DUTOK_UINT
        ;        
        
typeIdentifier:                  DUTOK_IDENTIFIER
        ;
        
booleanExpression:               expression
        ; 
        
ordinalTypedExpression:          expression
        ;             


%%

static void yyerror(unsigned *line, MlaskalCtx *, const char *m)
{
    error(DUERR_SYNTAX, *line, m);
}
